from spexm8p.token import Token

# 入力文字を、制御用として扱う文字。文字として扱う場合はエスケープが必要
ESCAPES = [
    Token.AND, Token.OR, Token.INVT,
    Token.CH_S, Token.CH_E, Token.WHOL, Token.DENY,
    Token.REPT,
    Token.SP_S, Token.SP_E,
]


def tokenize(input_str):
    ret = []
    escape_flg = False
    for char in list(input_str):
        if escape_flg:
            if char == Token.ESC:
                ret.append(Token.ESC + Token.ESC)
            elif char in ESCAPES:
                ret.append(Token.ESC + char)
            else:
                ret.append(Token.ESC)
                ret.append(char)
            escape_flg = False
        elif char == Token.ESC:
            escape_flg = True
        else:
            ret.append(char)
    if escape_flg:
        ret.append(Token.ESC + Token.ESC)
    return ret


def parse(tokens):
    return _parse_and_or(tokens)


def _parse_and_or(tokens):
    group_level = 0
    for i, token in enumerate(tokens):
        if token == Token.SP_S:
            group_level += 1
        elif token == Token.SP_E:
            group_level -= 1
        elif group_level == 0:
            if token == Token.AND:
                return _get_and_node(
                    _parse_and_or(tokens[:i]),
                    _parse_and_or(tokens[i+1:])
                )
            elif token == Token.OR:
                return _get_or_node(
                    _parse_and_or(tokens[:i]),
                    _parse_and_or(tokens[i+1:])
                )
    if group_level != 0:
        raise Exception(f'SyntaxError {Token.SP_E} not enough')
    return _parse_invert(tokens)


def _parse_invert(tokens):
    if len(tokens) == 0:
        raise Exception('SyntaxError invalid blank node')
    elif tokens[0] == Token.INVT:
        return _get_invert_node(_parse_concat(tokens[1:]))
    else:
        return _parse_concat(tokens)


def _parse_concat(tokens):
    group_level = 0
    in_ch = False
    current_tokens = []
    node_kind = 0  # 0: 登録不要, 1: group登録, 2: 単一文字登録, 3: 複数文字登録
    nodes = []
    for i, token in enumerate(tokens):
        if node_kind != 0:
            if node_kind == 1:
                node = _parse_and_or(current_tokens)
            elif node_kind == 2:
                node = _parse_inc_chex(current_tokens)
            elif node_kind == 3:
                node = _parse_chex(current_tokens)
            else:
                assert False  # unreachable code
            current_tokens = []
            node_kind = 0
            if token == Token.REPT:
                node = _get_repeat_node(node)
                nodes.append(node)
                continue
            nodes.append(node)
        if group_level == 0:
            if token == Token.INVT:
                raise Exception(f'SyntaxError {Token.INVT} Not at the beginning')
            elif token == Token.SP_E:
                raise Exception(f'SyntaxError {Token.SP_E} invalid position')
            elif token == Token.REPT:
                raise Exception(f'SyntaxError {Token.REPT} invalid position')
            elif token == Token.SP_S:
                if in_ch:
                    raise Exception(f'SyntaxError {Token.SP_S} invalid position')
                group_level += 1
            elif token == Token.CH_S:
                if in_ch:
                    raise Exception(f'SyntaxError {Token.CH_S} invalid position')
                in_ch = True
            elif token == Token.CH_E:
                if not in_ch:
                    raise Exception(f'SyntaxError {Token.CH_E} invalid position')
                in_ch = False
                node_kind = 3
            else:
                current_tokens.append(token)
                if not in_ch:
                    node_kind = 2
        else:
            if token == Token.SP_S:
                group_level += 1
                current_tokens.append(token)
            elif token == Token.SP_E:
                group_level -= 1
                if group_level != 0:
                    current_tokens.append(token)
                else:
                    node_kind = 1
            else:
                current_tokens.append(token)
    if group_level != 0:
        raise Exception(f'SyntaxError {Token.SP_E} not enough')
    if in_ch:
        raise Exception(f'SyntaxError {Token.CH_E} not enough')
    if node_kind != 0:
        if node_kind == 1:
            nodes.append(_parse_and_or(current_tokens))
        elif node_kind == 2:
            nodes.append(_parse_inc_chex(current_tokens))
        elif node_kind == 3:
            nodes.append(_parse_chex(current_tokens))
        else:
            assert False  # unreachable code
    if len(nodes) == 0:
        raise Exception('SyntaxError invalid blank node')
    elif len(nodes) == 1:
        return nodes[0]
    else:
        return _get_concat_node(nodes)


def _parse_chex(tokens):
    if len(tokens) > 0 and tokens[0] == Token.DENY:
        return _parse_exc_chex(tokens[1:])
    else:
        return _parse_inc_chex(tokens)


def _parse_inc_chex(tokens):
    if Token.WHOL in tokens:
        return _get_exc_chex([])
    else:
        return _get_inc_chex(tokens)


def _parse_exc_chex(tokens):
    if Token.WHOL in tokens:
        return _get_inc_chex([])
    else:
        return _get_exc_chex(tokens)


def _get_and_node(left, right):
    return {'type': 'and', 'left': left, 'right': right}


def _get_or_node(left, right):
    return {'type': 'or', 'left': left, 'right': right}


def _get_concat_node(nodes):
    return {'type': 'concat', 'nodes': nodes}


def _get_invert_node(node):
    return {'type': 'invert', 'node': node}


def _get_repeat_node(node):
    return {'type': 'repeat', 'node': node}


def _get_inc_chex(tokens):
    return {'type': 'inc_chex', 'tokens': tokens}


def _get_exc_chex(tokens):
    return {'type': 'exc_chex', 'tokens': tokens}



from functools import reduce
from spexm8p.spex import Chex
from spexm8p.spex import Spex
from spexm8p.parser import tokenize, parse


def spex(spex_str):
    return _build_spex(parse(tokenize(spex_str)))


def _build_spex(parsed):
    if parsed['type'] == 'inc_chex':
        return Spex.build_by_chex(Chex(parsed['tokens'], True))
    elif parsed['type'] == 'exc_chex':
        return Spex.build_by_chex(Chex(parsed['tokens'], False))
    elif parsed['type'] == 'or':
        return _build_spex(parsed['left']) | _build_spex(parsed['right'])
    elif parsed['type'] == 'and':
        return _build_spex(parsed['left']) & _build_spex(parsed['right'])
    elif parsed['type'] == 'invert':
        return ~_build_spex(parsed['node'])
    elif parsed['type'] == 'repeat':
        return _build_spex(parsed['node']).repeat()
    elif parsed['type'] == 'concat':
        return reduce(lambda a, b: a.concat(b), [_build_spex(x) for x in parsed['nodes']])
    else:
        assert False  # unreachable code



class Token:
    SP_S = '('  # 文字列集合_開始
    SP_E = ')'  # 文字列集合_終了
    AND = '&'   # 文字列集合_論理積
    OR = '|'    # 文字列集合_論理和
    INVT = '!'  # 文字列集合_否定
    REPT = '+'  # 文字列集合_1文字以上の繰返し
    CH_S = '['  # 文字集合_開始
    CH_E = ']'  # 文字集合_終了
    WHOL = '.'  # 文字集合_全集合
    DENY = '^'  # 文字集合_補集合
    ESC = '\\'  # エスケープ



from spexm8p.chex import Chex


# 文字列集合クラス (String Pattern Expression)
# 各状態をsid(state id)として保持する。
# sid =  0 は、開始状態として常に存在する前提とする。
class Spex:
    def __init__(self, tn_list, accepts_sids):
        self._accepts_sids = accepts_sids
        self._tns_dict = {}
        for tn in tn_list:
            fr_sid = tn._fr_sid
            if fr_sid not in self._tns_dict:
                self._tns_dict[fr_sid] = []
            self._tns_dict[fr_sid].append(tn)
        if len(self._accepts_sids) == 0:
            self._kind = 0  # 空集合
        elif len(self._accepts_sids) == len(self._tns_dict) - 1:
            self._kind = 1  # 全集合
        else:
            self._kind = 2

    # 空集合確認
    def blank(self):
        if self._kind == 0:
            return True
        else:
            return False

    # 全集合確認
    def whole(self):
        if self._kind == 1:
            return True
        else:
            return False

    # 否定論理演算(~)
    def __invert__(self):
        tn_list = []
        accepts_sids = set()
        for sid in self._tns_dict:
            if sid != 0 and sid not in self._accepts_sids:
                accepts_sids.add(sid)
            tn_list.extend(self._tns_dict[sid])
        return Spex(tn_list, accepts_sids)

    # 論理和演算(|)
    def __or__(self, other):
        if self._kind == 2:
            if other._kind == 2:
                new_tn_list = []
                new_accepts_sids = set()
                Spex._calc_and_or(
                    0, new_tn_list, new_accepts_sids,
                    self, other, 0, 0,
                    SidGen(), {}, 0,  # or
                )
                return Spex(new_tn_list, new_accepts_sids)
            elif other._kind == 0:
                return self
            elif other._kind == 1:
                return other
        elif self._kind == 0:
            return other
        elif self._kind == 1:
            return self

    # 論理積演算(&)
    def __and__(self, other):
        if self._kind == 2:
            if other._kind == 2:
                new_tn_list = []
                new_accepts_sids = set()
                Spex._calc_and_or(
                    0, new_tn_list, new_accepts_sids,
                    self, other, 0, 0,
                    SidGen(), {}, 1,  # and
                )
                return Spex(new_tn_list, new_accepts_sids)
            elif other._kind == 0:
                return other
            elif other._kind == 1:
                return self
        elif self._kind == 0:
            return self
        elif self._kind == 1:
            return other

    def concat(self, other):
        new_tn_list = []
        new_accepts_sids = set()
        Spex._calc_concat(
            0, new_tn_list, new_accepts_sids,
            self, other, 0, [],
            SidGen(), {},
        )
        return Spex(new_tn_list, new_accepts_sids)

    def repeat(self):
        new_tn_list = []
        new_accepts_sids = set()
        Spex._calc_repeat(
            0, new_tn_list, new_accepts_sids,
            self, [0],
            SidGen(), {}
        )
        return Spex(new_tn_list, new_accepts_sids)

    # 包含確認
    def include(self, other):
        if self._kind == 2:
            if other._kind == 2:
                return (~self & other).blank()
            elif other._kind == 0:
                return True
            elif other._kind == 1:
                return False
        elif self._kind == 0:
            return False
        elif self._kind == 1:
            return True

    def __str__(self):
        ret = ""
        for tns in self._tns_dict.values():
            for tn in tns:
                ret += f'{tn._fr_sid} -- {tn._chex} --> {tn._to_sid}\n'
        ret += f'[{", ".join([str(sid) for sid in self._accepts_sids])}]'
        return ret

    def __eq__(self, other):
        if self._kind == other._kind:
            if self._kind != 2:
                return True
            return self.include(other) and other.include(self)
        else:
            return False

    def mermaid(self):
        ret = "graph LR\n"
        for sid in self._tns_dict:
            if sid == 0:
                ret += "    0(( ))\n"
            elif sid == -1:
                ret += "    -1(( ))\n"
            elif sid == -2:
                ret += "    -2(( ))\n"
            else:
                ret += f"    {sid}( )\n"
        for sid in self._tns_dict:
            if sid == 0:
                ret += "    style 0 fill:#000,stroke-width:0px\n"
            elif sid == -1:
                ret += "    style -1 fill:#adb5bd,stroke-width:0px\n"
            elif sid == -2:
                ret += "    style -2 fill:#adb5bd,stroke:#dc3545,stroke-width:4px\n"
            elif sid in self._accepts_sids:
                ret += f"    style {sid} stroke:#dc3545,stroke-width:4px\n"
        for tns in self._tns_dict.values():
            for tn in tns:
                ret += f'    {tn._fr_sid} -- "{tn._chex}" --> {tn._to_sid}\n'
        return ret

    @classmethod
    def build_by_chex(cls, chex):
        if chex.blank():
            return Spex([
                Transition(0, 1, Chex.WHOLE),
                Transition(1, 1, Chex.WHOLE),
            ], set())
        else:
            return Spex([
                Transition(0, 1, chex),
                Transition(0, 2, ~chex),
                Transition(1, 2, Chex.WHOLE),
                Transition(2, 2, Chex.WHOLE),
            ], {1})

    @classmethod
    def build_whole(cls):
        return Spex([
            Transition(0, 1, Chex.WHOLE),
            Transition(1, 1, Chex.WHOLE)
        ], {1})

    @classmethod
    def _calc_and_or(cls, new_fr_sid, new_tn_list, new_accepts_sids, spex1, spex2, spex1_sid, spex2_sid, sid_gen, sid_dict_by_skey, ope_kind):
        # 1. 作成対象となる遷移先のパターンを抽出するため、chexを細分化する
        chex_pattern = [Chex.WHOLE]
        tmp_chex_pattern = []
        for target_chex in chex_pattern:
            for tn in spex1._tns_dict[spex1_sid]:
                and_chex = target_chex & tn._chex
                if not and_chex.blank():
                    tmp_chex_pattern.append(and_chex)
        chex_pattern = tmp_chex_pattern
        tmp_chex_pattern = []
        for target_chex in chex_pattern:
            for tn in spex2._tns_dict[spex2_sid]:
                and_chex = target_chex & tn._chex
                if not and_chex.blank():
                    tmp_chex_pattern.append(and_chex)
        chex_pattern = tmp_chex_pattern
        # 2. パターン毎に新しい遷移を追加する。 遷移先(new_to_sid)が新しい場合は再帰呼出を行う
        for target_chex in chex_pattern:
            # 2-1. 対象の遷移(next)をみつける
            for tn in spex1._tns_dict[spex1_sid]:
                if tn._chex.include(target_chex):
                    next_spex1_sid = tn._to_sid
                    break
            else:
                assert False  # unreachable code
            for tn in spex2._tns_dict[spex2_sid]:
                if tn._chex.include(target_chex):
                    next_spex2_sid = tn._to_sid
                    break
            else:
                assert False  # unreachable code
            # 2-2. 新しい遷移先のキーを作成する(再帰呼出済かの確認)
            skey = f'${next_spex1_sid}/${next_spex2_sid}'
            if skey in sid_dict_by_skey:
                # 2-2-1. 遷移先が登録済みの場合は、遷移先のsidを取得。(再帰呼出は行わない)
                new_to_sid = sid_dict_by_skey[skey]
            else:
                # 2-2-2. 遷移先が未登録の場合
                # 2-2-2-1. 新しい遷移先のsidを採番
                new_to_sid = sid_gen.get()
                # 2-2-2-2. 新しい遷移先を登録
                sid_dict_by_skey[skey] = new_to_sid
                # 2-2-2-3. 新しい遷移先の受理状態を評価(new_accepts_sids追加判定)
                if ope_kind == 0:  # or
                    if next_spex1_sid in spex1._accepts_sids or next_spex2_sid in spex2._accepts_sids:
                        new_accepts_sids.add(new_to_sid)
                elif ope_kind == 1:  # and
                    if next_spex1_sid in spex1._accepts_sids and next_spex2_sid in spex2._accepts_sids:
                        new_accepts_sids.add(new_to_sid)
                else:
                    assert False  # unreachable code
                # 2-2-2-4. 再帰呼出
                cls._calc_and_or(new_to_sid, new_tn_list, new_accepts_sids, spex1, spex2, next_spex1_sid, next_spex2_sid, sid_gen, sid_dict_by_skey, ope_kind)
            # 2-3. new_tn_list追加
            new_tn_list.append(Transition(new_fr_sid, new_to_sid, target_chex))

    @classmethod
    def _calc_concat(cls, new_fr_sid, new_tn_list, new_accepts_sids, spex1, spex2, spex1_sid, spex2_sids, sid_gen, sid_dict_by_skey):
        # 1. 作成対象となる遷移先のパターンを抽出するため、chexを細分化する
        chex_pattern = [Chex.WHOLE]
        tmp_chex_pattern = []
        for target_chex in chex_pattern:
            for tn in spex1._tns_dict[spex1_sid]:
                and_chex = target_chex & tn._chex
                if not and_chex.blank():
                    tmp_chex_pattern.append(and_chex)
        chex_pattern = tmp_chex_pattern
        if spex1_sid in spex1._accepts_sids:
            # spex1が受理状態の場合は、spex2の最初の遷移を考慮する
            tmp_chex_pattern = []
            for target_chex in chex_pattern:
                for tn in spex2._tns_dict[0]:
                    and_chex = target_chex & tn._chex
                    if not and_chex.blank():
                        tmp_chex_pattern.append(and_chex)
            chex_pattern = tmp_chex_pattern
        for spex2_sid in spex2_sids:
            tmp_chex_pattern = []
            for target_chex in chex_pattern:
                for tn in spex2._tns_dict[spex2_sid]:
                    and_chex = target_chex & tn._chex
                    if not and_chex.blank():
                        tmp_chex_pattern.append(and_chex)
            chex_pattern = tmp_chex_pattern
        # 2. パターン毎に新しい遷移を追加する。 遷移先(new_to_sid)が新しい場合は再帰呼出を行う
        for target_chex in chex_pattern:
            # 2-1. 対象の遷移(next)をみつける
            for tn in spex1._tns_dict[spex1_sid]:
                if tn._chex.include(target_chex):
                    next_spex1_sid = tn._to_sid
                    break
            else:
                assert False  # unreachable code
            next_spex2_sids = set()
            if spex1_sid in spex1._accepts_sids:
                # spex1が受理状態の場合は、spex2の最初の遷移を考慮する
                for tn in spex2._tns_dict[0]:
                    if tn._chex.include(target_chex):
                        next_spex2_sids.add(tn._to_sid)
                        break
                else:
                    assert False  # unreachable code
            for spex2_sid in spex2_sids:
                for tn in spex2._tns_dict[spex2_sid]:
                    if tn._chex.include(target_chex):
                        next_spex2_sids.add(tn._to_sid)
                        break
                else:
                    assert False  # unreachable code
            # 2-2. 新しい遷移先のキーを作成する(再帰呼出済かの確認)
            if len(next_spex2_sids) == 0:
                skey = f'${next_spex1_sid}/'
            else:
                skey = f'${next_spex1_sid}/${"-".join([str(x) for x in sorted(next_spex2_sids)])}'
            if skey in sid_dict_by_skey:
                # 2-2-1. 遷移先が登録済みの場合は、遷移先のsidを取得。(再帰呼出は行わない)
                new_to_sid = sid_dict_by_skey[skey]
            else:
                # 2-2-2. 遷移先が未登録の場合
                # 2-2-2-1. 新しい遷移先のsidを採番
                new_to_sid = sid_gen.get()
                # 2-2-2-2. 新しい遷移先を登録
                sid_dict_by_skey[skey] = new_to_sid
                # 2-2-2-3. 新しい遷移先の受理状態を評価(new_accepts_sids追加判定)
                for next_spex2_sid in next_spex2_sids:
                    if next_spex2_sid in spex2._accepts_sids:
                        new_accepts_sids.add(new_to_sid)
                        break
                # 2-2-2-4. 再帰呼出
                cls._calc_concat(new_to_sid, new_tn_list, new_accepts_sids, spex1, spex2, next_spex1_sid, next_spex2_sids, sid_gen, sid_dict_by_skey)
            # 2-3. new_tn_list追加
            new_tn_list.append(Transition(new_fr_sid, new_to_sid, target_chex))

    @classmethod
    def _calc_repeat(cls, new_fr_sid, new_tn_list, new_accepts_sids, spex, spex_sids, sid_gen, sid_dict_by_skey):
        # 1. 作成対象となる遷移先のパターンを抽出するため、chexを細分化する
        chex_pattern = [Chex.WHOLE]
        for spex_sid in spex_sids:
            tmp_chex_pattern = []
            for target_chex in chex_pattern:
                for tn in spex._tns_dict[spex_sid]:
                    and_chex = target_chex & tn._chex
                    if not and_chex.blank():
                        tmp_chex_pattern.append(and_chex)
            chex_pattern = tmp_chex_pattern
        # 受理状態の場合は、最初の遷移を考慮する
        for spex_sid in spex_sids:
            if spex_sid in spex._accepts_sids:
                tmp_chex_pattern = []
                for target_chex in chex_pattern:
                    for tn in spex._tns_dict[0]:
                        and_chex = target_chex & tn._chex
                        if not and_chex.blank():
                            tmp_chex_pattern.append(and_chex)
                chex_pattern = tmp_chex_pattern
                break
        # 2. パターン毎に新しい遷移を追加する。 遷移先(new_to_sid)が新しい場合は再帰呼出を行う
        for target_chex in chex_pattern:
            # 2-1. 対象の遷移(next)をみつける
            next_spex_sids = set()
            for spex_sid in spex_sids:
                if spex_sid in spex._accepts_sids:
                    # 受理状態の場合は、最初の遷移を考慮する
                    for tn in spex._tns_dict[0]:
                        if tn._chex.include(target_chex):
                            next_spex_sids.add(tn._to_sid)
                            break
                    break
            for spex_sid in spex_sids:
                for tn in spex._tns_dict[spex_sid]:
                    if tn._chex.include(target_chex):
                        next_spex_sids.add(tn._to_sid)
                        break
                else:
                    assert False  # unreachable code
            # 2-2. 新しい遷移先のキーを作成する(再帰呼出済かの確認)
            skey = "-".join([str(x) for x in sorted(next_spex_sids)])
            if skey in sid_dict_by_skey:
                # 2-2-1. 遷移先が登録済みの場合は、遷移先のsidを取得。(再帰呼出は行わない)
                new_to_sid = sid_dict_by_skey[skey]
            else:
                # 2-2-2. 遷移先が未登録の場合
                # 2-2-2-1. 新しい遷移先のsidを採番
                new_to_sid = sid_gen.get()
                # 2-2-2-2. 新しい遷移先を登録
                sid_dict_by_skey[skey] = new_to_sid
                # 2-2-2-3. 新しい遷移先の受理状態を評価(new_accepts_sids追加判定)
                for next_spex_sid in next_spex_sids:
                    if next_spex_sid in spex._accepts_sids:
                        new_accepts_sids.add(new_to_sid)
                        break
                # 2-2-2-4. 再帰呼出
                cls._calc_repeat(new_to_sid, new_tn_list, new_accepts_sids, spex, next_spex_sids, sid_gen, sid_dict_by_skey)
            # 2-3. new_tn_list追加
            new_tn_list.append(Transition(new_fr_sid, new_to_sid, target_chex))

    BLANK = None
    WHOLE = None


class Transition:
    def __init__(self, fr_sid, to_sid, chex):
        self._fr_sid = fr_sid
        self._to_sid = to_sid
        self._chex = chex


class SidGen:
    def __init__(self):
        self._sid = 0

    def get(self):
        self._sid += 1
        return self._sid


Spex.BLANK = Spex.build_by_chex(Chex.BLANK)
Spex.WHOLE = Spex.build_whole()



from spexm8p.token import Token


# 文字集合クラス (Character Expression)
class Chex:
    def __init__(self, chars, include_flg):
        self._char_set = set(chars)
        # self._include_flg => True: include, False: exclude(補集合)
        self._include_flg = True if include_flg else False
        self._len = len(self._char_set)
        if self._len == 0:
            if self._include_flg:
                self._kind = 0  # 空集合
            else:
                self._kind = 1  # 全集合
        else:
            self._kind = 2
        self._str = None  # 遅延評価

    # 空集合確認
    def blank(self):
        if self._kind == 0:
            return True
        else:
            return False

    # 全集合確認
    def whole(self):
        if self._kind == 1:
            return True
        else:
            return False

    # 否定論理演算(~)
    def __invert__(self):
        if self._kind == 2:
            return Chex(self._char_set, not self._include_flg)
        else:
            if self._kind == 0:
                return Chex.WHOLE
            elif self._kind == 1:
                return Chex.BLANK

    # 論理和演算(|)
    def __or__(self, other):
        if self._kind == 2:
            if other._kind == 2:
                if self._include_flg:
                    if other._include_flg:
                        return Chex(self._char_set | other._char_set, True)
                    else:
                        return Chex(other._char_set - self._char_set, False)
                else:
                    if other._include_flg:
                        return Chex(self._char_set - other._char_set, False)
                    else:
                        return Chex(self._char_set & other._char_set, False)
            elif other._kind == 0:
                return self
            elif other._kind == 1:
                return other
        elif self._kind == 0:
            return other
        elif self._kind == 1:
            return self

    # 論理積演算(&)
    def __and__(self, other):
        if self._kind == 2:
            if other._kind == 2:
                if self._include_flg:
                    if other._include_flg:
                        return Chex(self._char_set & other._char_set, True)
                    else:
                        return Chex(self._char_set - other._char_set, True)
                else:
                    if other._include_flg:
                        return Chex(other._char_set - self._char_set, True)
                    else:
                        return Chex(self._char_set | other._char_set, False)
            elif other._kind == 0:
                return other
            elif other._kind == 1:
                return self
        elif self._kind == 0:
            return self
        elif self._kind == 1:
            return other

    # 包含確認
    def include(self, other):
        if self._kind == 2:
            if other._kind == 2:
                if self._include_flg and not other._include_flg:
                    return False
                return (~self & other).blank()
            elif other._kind == 0:
                return True
            elif other._kind == 1:
                return False
        elif self._kind == 0:
            return False
        elif self._kind == 1:
            return True

    def __eq__(self, other):
        if self._kind == other._kind:
            if self._kind != 2:
                return True
            return str(self) == str(other)
        else:
            return False

    def __str__(self):
        if self._str is None:
            if self._kind == 2:
                chars = ''.join(sorted(self._char_set))
                if self._include_flg and self._len == 1:
                    self._str = chars
                else:
                    prefix = '' if self._include_flg else Token.DENY
                    self._str = f'{Token.CH_S}{prefix}{chars}{Token.CH_E}'
            elif self._kind == 0:
                self._str = f'{Token.CH_S}{Token.CH_E}'
            elif self._kind == 1:
                self._str = Token.WHOL
        return self._str

    BLANK = None
    WHOLE = None


Chex.BLANK = Chex([], True)
Chex.WHOLE = Chex([], False)
